// Harmless warning about browse information
#pragma warning (disable : 4786)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>

#include "Compress.h"

// Tweak values
const int gXPCompressionTweak1 = 5;
const int gXPCompressionTweak2 = 5;
const int gXPCompressionTweak3 = 5;
const int gXPCompressionTweak4 = 9;

#define FORCE_BYTE(x)						((unsigned char) (x))
#define FORCE_UNSIGNED(c)					c ## U

Compression::Compression()
{
}

Compression::~Compression()
{
}

void Compression::BitBufferWriteBits(void)
{
	u8 *p = mBitBufferCurrentOut;
	u32 b = mBitBufferValue;

	p[0] = FORCE_BYTE(b);
}


void Compression::BitBufferPutBit(unsigned char bit)
{
	if (mBitBufferCount < 8)
	{
		// Fetch new output pointer
		if (mBitBufferCount == 0)
		{
			mBitBufferCurrentOut = mBitBufferNextOut;
			mBitBufferNextOut++;
		}

		mBitBufferValue = (mBitBufferValue << 1) + bit;
		mBitBufferCount++;
	}
	else
	{
		BitBufferWriteBits();
		mBitBufferCurrentOut = mBitBufferNextOut;
		mBitBufferNextOut++;
		mBitBufferValue = bit;
		mBitBufferCount = 1;
	}
}


void Compression::BitBufferPutByte(unsigned char b)
{
	*mBitBufferNextOut++ = FORCE_BYTE(b);
}


void Compression::BitBufferFlushBits(unsigned char filler_bit)
{
	if (mBitBufferCount > 0)
	{
		while (mBitBufferCount != 8)
		{
			BitBufferPutBit(filler_bit);
		}
		BitBufferWriteBits();
		mBitBufferCount = 0;
	}
	mBitBufferCurrentOut = NULL;
}


void Compression::PackValue(u32 i)
{
	if (i >= 2)
	{
		u32 t = 4;
		i += 2;
		do
		{
			t = t << 1;
		} while (i >= t);
		t = t >> 1;
		do
		{
			t = t >> 1;
			BitBufferPutBit((i & t) ? 1 : 0);
			BitBufferPutBit(0);
		} while (t > 2);
	}
	BitBufferPutBit((unsigned)i & 1);
	BitBufferPutBit(1);
}


void Compression::EncodeMatch(u32 len, const u32 offset)
{
//	printf("EM%d,%d   ",len,offset);
	BitBufferPutBit(0);

	if (offset == mLastOffset)
	{
		BitBufferPutBit(0);
		BitBufferPutBit(1);
	}
	else
	{
		PackValue(1 + ((offset - 1) >> 8));
		BitBufferPutByte((unsigned)offset - 1);
	}
	len = len - 1;
	if (offset > LONG_OFFSET_THRESHOLD)
	{
		len--;
	}
	if (len >= 4)
	{
		BitBufferPutBit(0);
		BitBufferPutBit(0);
		PackValue(len - 4);
	}
	else
	{
		BitBufferPutBit(len > 1);
		BitBufferPutBit((unsigned)len & 1);
	}

	mLastOffset = offset;
}

void Compression::EncodeLiteralRun(const u8 *litRun, u32 currentLitNum)
{
	if (currentLitNum == 0)
	{
		return;
	}
//	printf("EL%d   ",currentLitNum);
	if (currentLitNum > mLastLit)
	{
		mLastLit = currentLitNum;
	}
	do
	{
		BitBufferPutBit(1);
		BitBufferPutByte(*litRun++);
		currentLitNum--;
	} while (currentLitNum > 0);
}


int Compression::CodedMatchLength(u32 len, u32 offset)
{
	int b;
	if (len < 2 || (len == 2 && (offset > LONG_OFFSET_THRESHOLD)))
		return -1;

	len = len - 2;
	if (offset > LONG_OFFSET_THRESHOLD)
	{
		len--;
	}

	if (offset == mLastOffset)
	{
		b = 1 + 2;
	}
	else
	{
		b = 1 + 10;
		offset = (offset - 1) >> 8;
		while (offset > 0)
		{
			b += 2;
			offset >>= 1;
		}
	}

	b += 2;
	if (len < 3)
	{
		return b;
	}
	len -= 3;

	do
	{
		b += 2;
		len >>= 1;
	} while (len > 0);

	return b;
}


int Compression::Compress( const u8 *in, u32 inLen,u8 *out, u32 *outLen,int compLevel)
{
	if ((compLevel < 1) || (compLevel > 10))
		return GBA_E_INVALID_ARGUMENT;

	mLastOffset = 1;
	mLastLit = 0;
	mInLen = inLen;
	mCurBuf = in;
	mBufEnd = in + inLen;

	mBitBufferValue = 0;
	mBitBufferCount = 0;
	mBitBufferCurrentOut = 0;
	mBitBufferNextOut = out;

	// Now the parts which decide compression speed and level
	mUpperLimitForwardMatchLen = compLevel * 1024;
	mMaximumRangeCheck = (compLevel * 32) / 11;
	mMaximumDictionaryBlocks = (compLevel * 131072) / 11;
	mDoExtendedDictionaryAdd = false;
	mDoForwardCheck = false;
	mUseDictionary = false;
	// MPi: TODO: For now we will always have the dictionary on even on the lowest compression.
	// It will be using only a small dictionary though.
//	if (compLevel > 1)		// Hence this "if" is commented out.
	{
		mUseDictionary = true;
	}
	if (compLevel > 6)
	{
		// Makes compression ~3 times slower than level 6
		mDoForwardCheck = true;
	}
	if (compLevel > 8)
	{
		// Makes compression ~2 times slower than level 8
		mDoExtendedDictionaryAdd = true;
	}
	// Special case for turning everything on to a sensible maximum
	if (compLevel == 10)
	{
		mUpperLimitForwardMatchLen = 0x7fffffff;
		mMaximumRangeCheck = 96;
		mMaximumDictionaryBlocks = 131072;
	}

	const u8 *litRun = mCurBuf;			// Pointer to the start of the literal run
	u32 currentLitNum = 0;			// Number of literals to encode

	u32 curPos = 0;

	mLastMatchPos = FORCE_UNSIGNED(-1);

	while (curPos < inLen)
	{
		int quickBits = -1;
		u32 proposedQuickMatchLen = 0;

		// Now try a dictionary match
		int dictionaryBits = -1;
		u32 matchLen,matchOffset;
		if (DictionaryFindMatch(curPos,matchLen,matchOffset))
		{
			dictionaryBits = CodedMatchLength(matchLen,matchOffset);
		}

		// Is length of the proposed match smaller than the number of bits needed for a literal version of the match length?
		if ((dictionaryBits != -1) && dictionaryBits < (matchLen * 9))
		{
			// Yes, so the match results in a smaller size so we use it.

			// Now do a forward match test
			u32 currentLitNum2 = currentLitNum + 1;
			u32 curPos2 = curPos+1;
			int dictionaryBits2 = -1;
			u32 matchLen2,matchOffset2;
			bool firstMatchIsBetter = true;
			// This forward match test takes some extra time
			if ((curPos2 < inLen) && mDoForwardCheck)
			{
				if (DictionaryFindMatch(curPos2,matchLen2,matchOffset2))
				{
					dictionaryBits2 = CodedMatchLength(matchLen2,matchOffset2);

					if ((dictionaryBits2 != -1) && dictionaryBits2 < (matchLen2 * 9))
					{
//						u32 forward = curPos2 - curPos;

						// Figure out if it is shorter (in terms of bits saved to the file) to skip the first match and encode using this second try instead
//						if ( (dictionaryBits + (int) ( ((forward * 9) + (matchLen2 - matchLen) * 9) / 2) ) > (dictionaryBits2 + (int)forward * 9))
//						if ( (dictionaryBits + (int) ( ((forward + matchLen2 - matchLen) * 9) / 2) ) > (dictionaryBits2 + (int)forward * 9))
//						if ( (dictionaryBits + (int) (int(forward) + int(matchLen2) - int(matchLen)) * 5) > (dictionaryBits2 + int(forward) * 9))
//						if ( (dictionaryBits + (int) (1 + int(matchLen2) - int(matchLen)) * 5) > (dictionaryBits2 + 9))
//						if ( (dictionaryBits + (int) 5 + (int(matchLen2) * 5) - (int(matchLen) * 5) ) > (dictionaryBits2 + 9))
						if ( (dictionaryBits + (int) gXPCompressionTweak1 + (int(matchLen2) * gXPCompressionTweak2) - (int(matchLen) * gXPCompressionTweak3) ) > (dictionaryBits2 + gXPCompressionTweak4))
						{
							firstMatchIsBetter = false;
						}
					}
				}
			}

			if (firstMatchIsBetter)
			{
				// Store any literals we have not already output
				EncodeLiteralRun(litRun,currentLitNum);
				// Store the match
				EncodeMatch(matchLen,matchOffset);
				curPos += matchLen;
				litRun = in + curPos;
				currentLitNum = 0;
//				printf("left = %d\n",inLen - curPos);
				continue;
			}
		}

		currentLitNum++;
		curPos++;
	}

	// Store any remaining literals
	EncodeLiteralRun(litRun,currentLitNum);

	// EOF
	BitBufferPutBit(0);
	PackValue(FORCE_UNSIGNED(0x1000000));
	BitBufferPutByte(0xff);
	BitBufferFlushBits(0);

	*outLen = (u32) (mBitBufferNextOut - out);

	// Makes sure the dictionary is freed
	DictionaryFree();
	return GBA_E_OK;
}

void Compression::DictionaryFree(void)
{
//	printf("mValueByBytes = %d\n",mValueByBytes.size());
	std::list<Value*>::iterator st,en;
	st = mByAge.begin();
	en = mByAge.end();
	while (st != en)
	{
		Value *value = *st++;
		// No need to unlink this value because we are clearing the containers anyway
		delete value;
	}
	mValueByBytes.clear();
	mByAge.clear();
	mValueByPos.clear();
}

void Compression::DictionaryAdd(const u32 fromStartPos)
{
	if ((fromStartPos+1) >= mInLen)
	{
		return;
	}

	// Quicker to do a find rather than allocating the Value block then failing on the first add
	if (mValueByPos.find(fromStartPos) != mValueByPos.end())
	{
		return;
	}

#if 1
	Bytes bytes(mCurBuf[fromStartPos],mCurBuf[fromStartPos+1]);
	Value *value = new Value;
	value->mPos = fromStartPos;
	// Fill in the iterators
	value->mByBytes = mValueByBytes.insert(std::pair<Bytes,Value *>(bytes,value));
	mByAge.push_back(value);
	value->mByAge = --mByAge.end();
	std::pair<std::map<u32,Value *>::iterator,bool> ret = mValueByPos.insert(std::pair<u32,Value *>(value->mPos,value));
	assert(ret.second && "XPCompression couldn't add the pair");
	value->mByPos = ret.first;

	// Expire old entries if we get too many
	// Saves memory at the expense of potential compression block matches
	// In most optimised STL implementation x.size() uses a counter variable so it isn't slow.
	if (mValueByBytes.size() > mMaximumDictionaryBlocks)
	{
		// Get the oldest used block from the head of the list
		Value *value = mByAge.front();
		DictionaryFreeValue(value);
	}
#endif
}

void Compression::DictionaryFreeValue(Value *value)
{
	mValueByBytes.erase(value->mByBytes);
	mByAge.erase(value->mByAge);
	mValueByPos.erase(value->mByPos);
	delete value;
}

bool Compression::DictionaryFindMatch(const u32 fromStartPos,u32 &len,u32 &matchOffset)
{
	u32 bestMatchPos = fromStartPos,bestMatchLen = 0;
	u32 maxNumBytesAhead = mInLen - fromStartPos;
	int bestBitLength = 0;

	// Have we searched for this already? If yes, early out.
	if (mLastMatchPos == fromStartPos)
	{
		len = mLastMatchLen;
		matchOffset = mLastMatchOffset;
		return mLastMatchReturn;
	}

	mLastMatchPos = fromStartPos;
	mLastMatchLen = 0;
	mLastMatchOffset = 0;
	mLastMatchReturn = false;

	DictionaryAdd(fromStartPos);

	if (fromStartPos == 0)
	{
		return false;
	}

#if 1
	// Try a simple backward search, slow but sure
	int pos;
	for (pos = fromStartPos - 1; (pos >= 0) && ((fromStartPos - pos) < mMaximumRangeCheck); pos--)
	{
		const u8 *start = mCurBuf + fromStartPos;
		const u8 *test = mCurBuf + pos;
		u32 i;
		u32 proposedMatchLen = 0;
		for (i=0; (i<mUpperLimitForwardMatchLen) && (i < maxNumBytesAhead);i++)
		{
			if (start[i] != test[i])
			{
				// This DictionaryAdd uses a bit more time and memory but improves the compression
				if (mDoExtendedDictionaryAdd)
				{
					DictionaryAdd(pos + i);
				}
				break;
			}
			proposedMatchLen = i+1;
		}
		if (proposedMatchLen == 0)
		{
			continue;
		}

		int proposedBits = CodedMatchLength(proposedMatchLen,fromStartPos - pos);
		// If we have a daft number of bits then don't use the match at all
		if (proposedBits == -1)
		{
			continue;
		}
		int proposedBitLength = (proposedMatchLen * 9) - proposedBits;

		if (proposedBitLength > bestBitLength)
		{
			bestMatchLen = proposedMatchLen;
			bestMatchPos = pos;
			bestBitLength = proposedBitLength;
		}
		// Logic end begins here
	}
#endif

#if 1
	if (mUseDictionary && (maxNumBytesAhead >= 2) && (bestMatchLen < maxNumBytesAhead))
	{
		// Search the current dictionary for byte sequences that might match what we want
		Bytes bytes(mCurBuf[fromStartPos],mCurBuf[fromStartPos+1]);
		std::multimap<Bytes,Value *,ByBytes>::iterator st = mValueByBytes.lower_bound(bytes);
		std::multimap<Bytes,Value *,ByBytes>::iterator en = mValueByBytes.upper_bound(bytes);
//		printf(".\n");
		// MPi: TODO: Check how often we get collisions with the typical data and optimise for needed
		while (st != en)
		{
//			printf(".");
			// Check the byte sequences that are found
			Value *value = (*st++).second;
			u32 pos = value->mPos;
			// Stops us from comparing with ourselves
			if ((fromStartPos - pos) < mMaximumRangeCheck)
			{
				continue;
			}

			// Logic begins here
			const u8 *start = mCurBuf + fromStartPos;
			const u8 *test = mCurBuf + pos;
			bool used = false;
			u32 i;
			u32 proposedMatchLen = 0;
			for (i=0; (i<mUpperLimitForwardMatchLen) && (i < maxNumBytesAhead);i++)
			{
				if (start[i] != test[i])
				{
					break;
				}
				proposedMatchLen = i+1;
			}
			if (proposedMatchLen == 0)
			{
				continue;
			}

			int proposedBits = CodedMatchLength(proposedMatchLen,fromStartPos - pos);
			// If we have a daft number of bits then don't use the match at all
			if (proposedBits == -1)
			{
				continue;
			}
			int proposedBitLength = (proposedMatchLen * 9) - proposedBits;

			if (proposedBitLength > bestBitLength)
			{
				bestMatchLen = proposedMatchLen;
				bestMatchPos = pos;
				bestBitLength = proposedBitLength;
				used = true;
			}

			if (used)
			{
				// Because this entry is used it is put back at the end of the list again
				mByAge.erase(value->mByAge);
				mByAge.push_back(value);
				value->mByAge = --mByAge.end();
			}

		}
	}
#endif

	if (bestMatchLen == 0)
	{
		return false;
	}

	// Do the forward match test with whatever the match is and calculate a new length based on how many times the sequence repeats
	u32 to = mInLen - (fromStartPos + bestMatchLen);
	const u8 *start = mCurBuf + fromStartPos + bestMatchLen;
	const u8 *test = mCurBuf + bestMatchPos + bestMatchLen;
	while ((start < mBufEnd) /*&& (bestMatchLen < mUpperLimitForwardMatchLen)*/)
	{
		if (*start != *test)
		{
			break;
		}
		start++;
		test++;
		bestMatchLen++;
	}

	// Now finally return the best match
	matchOffset = fromStartPos - bestMatchPos;
	len = bestMatchLen;

	mLastMatchLen = len;
	mLastMatchOffset = matchOffset;
	mLastMatchReturn = true;

	return true;
}
